package pl.edu.agh.kis.eis.ndl

import scala.annotation.tailrec
import scala.collection.SortedSet
import scala.util.Random

class KnowledgeBase[A](facts: Set[Product])(implicit ordering: Ordering[A]) {
  type Fact1 = Tuple1[A]
  type Fact2 = (A, A)

  // Term indexes
  val facts1: SortedSet[A] = SortedSet(facts.collect { case f: Fact1 => f._1 }.toSeq: _*)
  val fact2Index1: Map[A, SortedSet[A]] = facts.collect { case f: Fact2 => f }
    .groupBy(_._1).map { case (x, y) => (x, SortedSet(y.map(_._2).toSeq: _*)) }
  val fact2Index2: Map[A, SortedSet[A]] = facts.collect { case f: Fact2 => f }
    .groupBy(_._2).map { case (x, y) => (x, SortedSet(y.map(_._1).toSeq: _*)) }

  val valueSetIteratorPartial: PartialFunction[Placeholder[A], (String, BufferedIterator[A])] = {
    case Placeholder(Const(a), v: Variable[A]) => (v.name.get, fact2Index1(a).iterator.buffered)
    case Placeholder(v: Variable[A], Const(a)) => (v.name.get, fact2Index2(a).iterator.buffered)
    case Placeholder(v: Variable[A]) => (v.name.get, facts1.iterator.buffered)
  }

  def minNextValueIterator(x: Seq[BufferedIterator[A]]): Option[Int] = {
    val vals = x.zipWithIndex.map { case (u, index) => if (u.hasNext) Some(index, u.head) else None }
    if (vals.forall(_.isEmpty)) return None
    val minIndex = vals.collect { case Some(z) => z }.minBy(_._2)
    Some(minIndex._1)
  }

  def allNonEqual(x: Iterable[_]): Boolean = x.exists(_ != x.head)

  @tailrec
  final def nextVarSolution(prevValues: Seq[Option[A]], iterators: Seq[BufferedIterator[A]]): Option[A] = {
    val minIteratorIndex = minNextValueIterator(iterators)
    if (minIteratorIndex.isEmpty) return None
    val updatedValues = prevValues.updated(minIteratorIndex.get, Some(iterators(minIteratorIndex.get).next()))

    if (allNonEqual(updatedValues)) { // at least one value is different
      nextVarSolution(updatedValues, iterators)
    } else {
      // all equal
      if (updatedValues.head.isDefined)
        Some(updatedValues.head.get) // all defined and equal
      else
        None // all None, search space exhausted
    }
  }

  def variableIterator(variable: String, placeholders: Seq[Placeholder[A]]): Iterable[Variable[A]] = {
    val iterators: Seq[BufferedIterator[A]] = placeholders
      .filter(_.variables.exists(_.name.contains(variable)))
      .collect(valueSetIteratorPartial).map(_._2)

    new Iterable[Variable[A]] {
      override def iterator: Iterator[Variable[A]] = new Iterator[Variable[A]] {
        var elem: Option[Variable[A]] = None
        var elemValid = false

        private[this] def update(): Unit = {
          elem = nextVarSolution(Seq.tabulate(iterators.length)(_ => None), iterators).map { value =>
            Variable(variable, value)
          }
          elemValid = true
        }

        override def hasNext: Boolean = {
          if (!elemValid) update()
          elem.nonEmpty
        }

        override def next(): Variable[A] = {
          if (!elemValid) update()
          val res = elem.get
          elemValid = false
          res
        }
      }
    }
  }

  def crossJoin[T](list: Traversable[Iterable[T]]): Traversable[Iterable[T]] =
    list match {
      case xs :: Nil => xs map (Iterable(_))
      case x :: xs => for {
        i <- x
        j <- crossJoin(xs)
      } yield Iterable(i) ++ j
    }

  def queryAll(q: Placeholder[A]*): Iterator[Set[Variable[A]]] = {
    val variables: Set[String] = q.flatMap(_.variables.map(_.name.get)).toSet
    val variableIterators = variables.map(variableIterator(_, q)).map(_.toStream.toIterable).toList
    val solutionIterator = crossJoin(variableIterators).toIterator

    new Iterator[Set[Variable[A]]] {
      override def hasNext: Boolean = solutionIterator.hasNext
      override def next(): Set[Variable[A]] = solutionIterator.next().toSet
    }
  }

  def queryRandom(q: Placeholder[A]*): Option[Set[Variable[A]]] = {
    val randomSolution = Random.shuffle(queryAll(q: _*))
    if (randomSolution.hasNext) Some(randomSolution.next()) else None
  }
}
