package pl.edu.agh.kis.eis.ndl

import org.scalatest.{FlatSpec, Matchers}

class KnowledgeBaseSpec extends FlatSpec with Matchers {
  "Facts" should "be indexed" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      Tuple1(1), Tuple1(2), Tuple1(3),
      (1, 2), (2, 3), (2, 1), (3, 3)
    ))

    kb.facts1 should contain only (1, 2, 3)
    kb.fact2Index1 should contain only (
      1 -> Set(2),
      2 -> Set(3, 1),
      3 -> Set(3)
    )
    kb.fact2Index2 should contain only (
      2 -> Set(1),
      3 -> Set(2, 3),
      1 -> Set(2)
    )
  }

  /*
   * Test cases use  "queryAll(...).toSeq should contain theSameElementsAs" to check whether:
   * a) an iterator is finite,
   * b) does not repeat solutions
   *
   * FIXME: create test-case template and add this note there
   */

  "Query ((x) (x 3)) #1" should "be unified" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      Tuple1(1), Tuple1(2), Tuple1(3),
      (1, 2), (2, 3), (2, 1), (3, 3)
    ))

    val x = Variable[Int]("x")
    kb.queryAll(Placeholder(x), Placeholder(x, Const(3))).toSeq should contain theSameElementsAs Seq(
      Set(Variable("x", 2)),
      Set(Variable("x", 3))
    )
  }

  "Query ((x))" should "return all unary facts" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(Tuple1(1), Tuple1(2)))

    kb.queryAll(Placeholder(Variable("x"))).toSeq should contain theSameElementsAs Seq(
      Set(Variable("x", 1)),
      Set(Variable("x", 2))
    )
  }

  "Query ((x) (x 3))" should "be unified" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      Tuple1(8), Tuple1(3), Tuple1(1), Tuple1(2),
      (1, 1), (1, 2), (1, 8), (2, 1), (4, 1), (8, 1), (3, 8), (8, 3)
    ))

    val x = Variable[Int]("x")
    kb.queryAll(Placeholder(x), Placeholder(x, Const(3))).toSeq should contain theSameElementsAs Seq(
      Set(Variable("x", 8))
    )

    kb.queryAll(Placeholder(x), Placeholder(x, Const(1))).toSeq should contain theSameElementsAs Seq(
      Set(Variable("x", 1)),
      Set(Variable("x", 2)),
      Set(Variable("x", 8))
    )
  }

  "Query ((x) (y))" should "be unified" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      Tuple1(1), Tuple1(14)
    ))

    kb.queryAll(Placeholder(Variable("x")), Placeholder(Variable("y"))).toSeq should contain theSameElementsAs Seq(
      Set(Variable("x", 1), Variable("y", 1)),
      Set(Variable("x", 1), Variable("y", 14)),
      Set(Variable("x", 14), Variable("y", 1)),
      Set(Variable("x", 14), Variable("y", 14))
    )
  }

  "Query ((x) (y) (x 1) (y 2))" should "unify variables" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      Tuple1(1), Tuple1(2),
      (1, 1), (1, 2), (2, 1)
    ))

    val x = Variable[Int]("x")
    val y = Variable[Int]("y")

    kb.queryAll(
      Placeholder(x), Placeholder(y), Placeholder(x, Const(1)), Placeholder(y, Const(2))
    ).toSeq should contain theSameElementsAs Seq(
      Set(Variable("y", 1), Variable("x", 1)),
      Set(Variable("y", 1), Variable("x", 2))
    )
  }

  // FIXME take many samples and check all solutions existence - currently may success with part of solutions
  it should "unify variables randomly" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      Tuple1(1), Tuple1(2),
      (1, 1), (1, 2), (2, 1)
    ))

    val x = Variable[Int]("x")
    val y = Variable[Int]("y")

    kb.queryRandom(
      Placeholder(x), Placeholder(y), Placeholder(x, Const(1)), Placeholder(y, Const(2))
    ) should contain oneOf (
      Set(Variable("y", 1), Variable("x", 2)),
      Set(Variable("y", 1), Variable("x", 3))
    )
  }

  "Query ((x) (x 1))" should "not unify variables" in {
    val kb: KnowledgeBase[Int] = new KnowledgeBase(Set(
      (1, 1), (1, 2), (1, 8), (2, 1), (4, 1), (8, 1), (3, 8), (8, 3)
    ))
    val x = Variable[Int]("x")
    kb.queryAll(Placeholder(x), Placeholder(x, Const(1))).toSeq should be(empty)
  }

  "Query ((x) (y) (x y))" should "be unified" ignore {
    val kb: KnowledgeBase[Int] = new KnowledgeBase[Int](Set(
      Tuple1(2), Tuple1(8), (2, 8), // solution: x = 2, y = 8
      Tuple1(5), Tuple1(9), (9, 5), // solution: x = 9, y = 5
      Tuple1(3), Tuple1(4) // not a solution, missing (3, 4) or (4, 3)
    ))

    val result = kb.queryAll(
      Placeholder(Variable("x")),
      Placeholder(Variable("y")),
      Placeholder(Variable("x"), Variable("y"))).toSet

    result should contain only (
      Set(Variable("x", 2), Variable("y", 8)),
      Set(Variable("x", 9), Variable("y", 5))
    )
  }

  "Query ((x 1) (y 2))" should "be unified" ignore {
  }

  "Query ((x) (y) (z z))" should "be unified" ignore {
  }
}
